using System;
using System.Collections.Generic;
using System.Drawing;
using System.Threading;
using System.Windows.Forms;
using P4Lib;

/*************************************************************************************
 *
 *  P4Ctrl
 *  Author : Bidou (http://www.csharpfr.com/auteurdetail.aspx?ID=13319)
 *  Blog   : http://blogs.developpeur.org/bidou/
 *  Date   : February 2007
 * 
 *************************************************************************************/

namespace P4Ctrl
{
    /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    /// <summary>
    /// Represents the game (UserControl).
    /// </summary>
    /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    public partial class Grid : UserControl
    {
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Handler used when a player win.
        /// </summary>
        /// <param name="playerType"> The winner. </param>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public delegate void UserWinEventHandler(PlayerType playerType);
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Fired when a user win.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public event UserWinEventHandler UserWin = null;
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Fired when the computer starts to think.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public event EventHandler ComputerStartsThinking = null;
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Fired when the computer ends to think.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public event EventHandler ComputerEndsThinking = null;

        private bool _isThinking = false;
        private bool _isGameFinished = false;
        private int _squareSide = 60;
        private Thread _threadComputer = null;
        private Color _humanColor = Color.Red;
        private List<Point> _winnerPts = new List<Point>(4);
        private Game _game = null;
        private AI _ai = null;

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Create the UserControl.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public Grid()
        {
            this.InitializeComponent();
            this.Restart(false, Difficulty.Hard);
            this._isGameFinished = true;
        }

        #region Properties

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Get or set the size of a square.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public int SquareSide
        {
            get { return this._squareSide; }
            set { this._squareSide = value; }
        }
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Get or set the human's color.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public Color HumanColor
        {
            get { return this._humanColor; }
            set { this._humanColor = value; }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Get if the game is finished.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public bool IsFinished
        {
            get { return this._isGameFinished; }
        }

        #endregion

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Start or restart the game.
        /// </summary>
        /// <param name="computerPlayFirst"> True if the computer plays first, otherwise false. </param>
        /// <param name="level"> The difficulty. </param>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public void Restart(bool computerPlayFirst, Difficulty level)
        {
            this.Stop();
            this._game = new Game(computerPlayFirst);
            this._game.HasPlayed += new EventHandler(this.game_HasPlayed);
            this._ai = new AI(level);
            this._winnerPts = new List<Point>(4);
            this._isGameFinished = false;

            if (computerPlayFirst) this.PlayComputerAsyc();
            this.ClientSize = new Size(this._game.Width * this._squareSide, this._game.Height * this._squareSide);
            this.Refresh();
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Stop the current game.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public void Stop()
        {
            if (this._threadComputer != null)
            {
                this._threadComputer.Abort();
                this._threadComputer = null;
                this._isThinking = false;
                if (this.ComputerEndsThinking != null) this.ComputerEndsThinking(this, EventArgs.Empty);
            }
            this._isGameFinished = true;
            this.Invalidate();
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Raised when the computer has played.
        /// </summary>
        /// <param name="sender"> Not used. </param>
        /// <param name="e"> Not used. </param>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        private void game_HasPlayed(object sender, EventArgs e)
        {
            this._isThinking = false;
            if (this.ComputerEndsThinking != null) this.ComputerEndsThinking(this, EventArgs.Empty);
            if (!this._game.CanPlay())
            {
                this._isGameFinished = true;
                if (this.UserWin != null) this.UserWin(PlayerType.NoBody);
            }
            else if (this._game.IsWinner(PlayerType.Computer, out this._winnerPts))
            {
                this._isGameFinished = true;
                if (this.UserWin != null) this.UserWin(PlayerType.Computer);
                this.Invalidate();
            }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Draw the game.
        /// </summary>
        /// <param name="e"></param>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            using (Brush game = new SolidBrush(Color.Black))
            using (Brush ellipseEmpty = new SolidBrush(this._isGameFinished ? Color.Gray : Color.White))
            using (Brush ellipseHuman = new SolidBrush(this._humanColor))
            using (Brush ellipseComputer = new SolidBrush(Color.Green))
            using (Pen border = new Pen(Color.Gray, 3))
            {
                for (int i = 0; i < this._game.Width; i++)
                {
                    for (int j = 0; j < this._game.Height; j++)
                    {
                        Rectangle rectangle = new Rectangle(i * this._squareSide, j * this._squareSide, this._squareSide, this._squareSide);
                        e.Graphics.FillRectangle(game, rectangle);
                        e.Graphics.DrawRectangle(border, rectangle);
                        rectangle.Inflate(-7, -7);
                        // The circle
                        switch(this._game[i, j].PlayerType)
                        {
                            case PlayerType.NoBody: e.Graphics.FillEllipse(ellipseEmpty, rectangle); break;
                            case PlayerType.Computer: e.Graphics.FillEllipse(ellipseComputer, rectangle); break;
                            case PlayerType.Human: e.Graphics.FillEllipse(ellipseHuman, rectangle); break;
                        }
                    }
                }
            }
            // The winner
            using (Pen winnerPen = new Pen(Color.White, 2f))
            {
                foreach (Point winnerPt in this._winnerPts)
                {
                    Rectangle f = this.PointToRectangle(winnerPt);
                    f.Inflate(-8, -8);
                    e.Graphics.DrawEllipse(winnerPen, f);
                }
            }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// The user click on the grid.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        private void Grid_MouseClick(object sender, MouseEventArgs e)
        {
            if (!this._isGameFinished && !this._isThinking)
            {
                int column = this.PointToColumn(e.Location);
                if (this._game.CanPlay(column))
                {
                    this.PlayHuman(column);

                    if (this._game.IsWinner(PlayerType.Human, out this._winnerPts))
                    {
                        this._isGameFinished = true;
                        if (this.UserWin != null) this.UserWin(PlayerType.Human);
                    }
                    else if (!this._game.CanPlay())
                    {
                        this._isGameFinished = true;
                        if (this.UserWin != null) this.UserWin(PlayerType.NoBody);
                    }
                    else this.PlayComputerAsyc();
                }
                if (this._isGameFinished) this.Invalidate();
            }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Play (human).
        /// </summary>
        /// <param name="column"> The column to play to. </param>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        private void PlayHuman(int column)
        {
            Point playPlayerPt = this._game.Play(column, Player.Human);
            this.Invalidate(this.PointToRectangle(playPlayerPt));
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Play in a thread (computer).
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        private void PlayComputerAsyc()
        {
            if (this.ComputerStartsThinking != null) this.ComputerStartsThinking(this, EventArgs.Empty);
            this._isThinking = true;
            this._threadComputer = new Thread(new ThreadStart(this.PlayComputer));
            this._threadComputer.Start();
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Play (computer).
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        private void PlayComputer()
        {
            Point playComputerPt = this._game.Play(this._ai.GetBestColumn(this._game), Player.Computer);
            this.Invalidate(this.PointToRectangle(playComputerPt));
        }

        #region Utilities

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Convert a point to a column.
        /// </summary>
        /// <param name="pt"> The point to convert. </param>
        /// <returns> The rectangle. </returns>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        private int PointToColumn(Point pt)
        {
            return (pt.X / this._squareSide);
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Convert a point to a rectangle.
        /// </summary>
        /// <param name="pt"> The point to convert. </param>
        /// <returns> The rectangle. </returns>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        private Rectangle PointToRectangle(Point pt)
        {
            return new Rectangle(pt.X * this._squareSide, pt.Y * this._squareSide, this._squareSide, this._squareSide);
        }

        #endregion

    }
}