using System;
using System.Collections.Generic;
using System.Drawing;

/*************************************************************************************
 *
 *  P4Lib
 *  Author : Bidou (http://www.csharpfr.com/auteurdetail.aspx?ID=13319)
 *  Blog   : http://blogs.developpeur.org/bidou/
 *  Date   : February 2007
 * 
 *************************************************************************************/

namespace P4Lib
{
    /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    /// <summary>
    /// Represents the game.
    /// </summary>
    /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    public class Game
    {
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Reached when the computer has played.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public event EventHandler HasPlayed = null;

        private int _bestColumnToPlayTo = -1;
        private int _gameValue = 0;
        private int _levelInTree = 0;
        private Player[,] _players = new Player[7, 6];
        private PlayerType _currentPlayer = PlayerType.NoBody;

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Create a new game.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public Game(bool computerPlaysFirst)
        {
            this._currentPlayer = computerPlaysFirst ? PlayerType.Human : PlayerType.Computer;
            this.InitGame();
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Create a new game with the specified positions.
        /// </summary>
        /// <param name="players"> The player's positions. </param>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        private Game(Player[,] players)
        {
            this._players = players;
        }

        #region Properties

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// The best column for the current game.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public int BestColumn
        {
            get { return this._bestColumnToPlayTo; }
            set { this._bestColumnToPlayTo = value; }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Get or set the player.
        /// </summary>
        /// <param name="x"> The x-coordinate. </param>
        /// <param name="y"> The y-coordinate. </param>
        /// <returns> The player at the specified location. </returns>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public Player this[int x, int y]
        {
            get { return this._players[x, y]; }
            internal set { this._players[x, y] = value; }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Get or set the player.
        /// </summary>
        /// <param name="pt"> The location of the player. </param>
        /// <returns> The player at the specified location. </returns>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        private Player this[Point pt]
        {
            get { return this[pt.X, pt.Y]; }
            set { this[pt.X, pt.Y] = value; }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Get the level of the game in the tree.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public int LevelInTree
        {
            get { return this._levelInTree; }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Get if the current game is a leaf.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public bool IsLeaf
        {
            get
            {
                bool isLeaf = true;
                List<Point> notUsed = null;
                // State of the game
                if (!this.CanPlay()) this._gameValue = 0;
                else if (this.IsWinner(PlayerType.Computer, out notUsed)) this._gameValue = 1;
                else if (this.IsWinner(PlayerType.Human, out notUsed)) this._gameValue = -1;
                else
                {
                    this._gameValue = 0;
                    isLeaf = false;
                }
                return isLeaf;
            }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// The current player.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public PlayerType CurrentPlayer
        {
            get { return this._currentPlayer; }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Get or set the value of the current game.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public int Value
        {
            get { return this._gameValue; }
            private set { this._gameValue = value; }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Get the width of the game.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public int Width
        {
            get { return this._players.GetLength(0); }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Get the height of the game.
        /// </summary>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public int Height
        {
            get { return this._players.GetLength(1); }
        }

        #endregion

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// A Player plays (specified column).
        /// </summary>
        /// <param name="column"> The columns to play to. </param>
        /// <param name="player"> The player which plays. </param>
        /// <returns> The point where the player has played. </returns>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public Point Play(int column, Player player)
        {
            Point pt = new Point(column, this.GetRow(column));
            this[pt] = player;
            if (player.PlayerType == PlayerType.Computer && this.HasPlayed != null) this.HasPlayed(this, EventArgs.Empty);
            return pt;
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Check if the game is full.
        /// </summary>
        /// <returns> False if the game is full, otherwise true. </returns>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public bool CanPlay()
        {
            for (int i = 0; i < this.Width; i++)
            {
                if (this.CanPlay(i)) return true;
            }
            return false;
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// Check if a column is full.
        /// </summary>
        /// <param name="column"> The column to check. </param>
        /// <returns> False if the column is full, otherwise true. </returns>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public bool CanPlay(int column)
        {
            return ((column >= 0 && column < this.Width) ? this._players[column, 0].PlayerType == PlayerType.NoBody : false);
        }

        /// --------------------------------------------------------------------------------------------
        /// <summary>
        /// Check if a point is valid.
        /// </summary>
        /// <param name="pt"> The point to check. </param>
        /// <returns> True if the point is valid, otherwise false. </returns>
        /// --------------------------------------------------------------------------------------------
        public bool IsPointValid(Point pt)
        {
            return (pt.X >= 0 && pt.X < this.Width && pt.Y >= 0 && pt.Y < this.Height);
        }

        /// --------------------------------------------------------------------------------------------
        /// <summary>
        /// Check if a player wins.
        /// </summary>
        /// <param name="player"> The player to check. </param>
        /// <param name="winnerPts"> The 4 winner-points.</param>
        /// <returns> True if the player wins. </returns>
        /// --------------------------------------------------------------------------------------------
        public bool IsWinner(PlayerType winner, out List<Point> winnerPts)
        {
            winnerPts = new List<Point>();
            for (int i = 0; i < this.Width; i++)
            {
                for (int j = 0; j < this.Height; j++)
                {
                    Point pt = new Point(i, j);
                    int x = -2;
                    int y = -2;

                    if (this[pt].PlayerType != winner) continue;
                    // Check directions
                    if (this.CheckDir(pt, 0, 1)) { x = 0; y = 1; }
                    else if (this.CheckDir(pt, 1, 0)) { x = 1; y = 0; }
                    else if (this.CheckDir(pt, 0, -1)) { x = 0; y = -1; }
                    else if (this.CheckDir(pt, -1, 0)) { x = -1; y = 0; }
                    else if (this.CheckDir(pt, 1, 1)) { x = 1; y = 1; }
                    else if (this.CheckDir(pt, 1, -1)) { x = 1; y = -1; }
                    else if (this.CheckDir(pt, -1, 1)) { x = -1; y = 1; }
                    else if (this.CheckDir(pt, -1, -1)) { x = -1; y = -1; }

                    if (x > -2 && y > -2)
                    {
                        for (int k = 0; k < 4; k++) winnerPts.Add(new Point(pt.X + x * k, pt.Y + y * k));
                        return true;
                    }
                }
            }
            return false;
        }

        /// --------------------------------------------------------------------------------------------
        /// <summary>
        /// Check in a direction if there are 4 pawns.
        /// </summary>
        /// <param name="pt"> The start point. </param>
        /// <param name="x"> The x-coordinate. </param>
        /// <param name="y"> The y-coordinate. </param>
        /// <returns> True if there is 4 onsecutive points. </returns>
        /// --------------------------------------------------------------------------------------------
        private bool CheckDir(Point pt, int x, int y)
        {
            PlayerType potWinner = this[pt].PlayerType;
            if (potWinner == PlayerType.NoBody) return false;
            else
            {
                for (int i = 1; i < 4; i++)
                {
                    Point toPlay = new Point(pt.X + x * i, pt.Y + y * i);
                    if (!IsPointValid(toPlay) || this[toPlay].PlayerType != potWinner) return false;
                }
            }
            return true;
        }

        /// --------------------------------------------------------------------------------------------
        /// <summary>
        /// Get the row where the pawn fell if you play in 'column'.
        /// </summary>
        /// <param name="column"> The column. </param>
        /// <returns> The row. </returns>
        /// --------------------------------------------------------------------------------------------
        private int GetRow(int column)
        {
            int rows = -1;
            for (int i = 0; i < this.Height; i++)
            {
                if (this._players[column, i].PlayerType != PlayerType.NoBody)
                {
                    rows = i - 1;
                    break;
                }
            }
            if (rows < 0) rows = this.Height - 1;
            return rows;
        }

        /// --------------------------------------------------------------------------------------------
        /// <summary>
        /// Initialisation of the game.
        /// </summary>
        /// --------------------------------------------------------------------------------------------
        private void InitGame()
        {
            for (int i = 0; i < this.Width; i++)
            {
                for (int j = 0; j < this.Height; j++)
                {
                    this._players[i, j] = new Player(PlayerType.NoBody);
                }
            }
        }

        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /// <summary>
        /// The possible successors for a game.
        /// </summary>
        /// <param name="player"> The player. </param>
        /// <returns> A collection of games. </returns>
        /// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        public List<Game> GetSuccessors(PlayerType player)
        {
            List<Game> nextGame = new List<Game>();
            for (int i = 0; i < this.Width; i++)
            {
                if (this.CanPlay(i))
                {
                    Player[,] players = this._players.Clone() as Player[,];     // Clone the current game
                    players[i, this.GetRow(i)].PlayerType = player;             // The new location  
                    Game newGame = new Game(players);                           // The successor
                    newGame._bestColumnToPlayTo = i;                            // The best column
                    newGame._levelInTree++;                                     // The level in the tree
                    newGame._currentPlayer = player == PlayerType.Computer ? PlayerType.Human : PlayerType.Computer;            // Change the current player
                    nextGame.Add(newGame);                                      // Add to the successor collection
                }
            }
            return nextGame;
        }

        #region NotUsed

        public IEnumerable<Player> Pawns
        {
            get
            {
                for (int i = 0; i < this.Width; i++)
                {
                    for (int j = 0; j < this.Height; j++)
                    {
                        yield return this._players[i, j];
                    }
                }
            }
        }

        #endregion
    }
}